require('vis')

local m = vis.modes
local mode_strings = {
	[m.NORMAL]           = 'NORMAL',
	[m.OPERATOR_PENDING] = 'OPERATOR-PENDING',
	[m.VISUAL]           = 'VISUAL',
	[m.VISUAL_LINE]      = 'VISUAL-LINE',
	[m.INSERT]           = 'INSERT',
	[m.REPLACE]          = 'REPLACE',
}

local e = vis.events

-- init
e.subscribe(e.INIT, function()
	-- load theme
	require("themes/basic")
	local lexers = vis.lexers
	lexers.lexers = {}
	if lexers.load and not lexers.property then lexers.load("text") end

	-- mappings
	vis:map(m.NORMAL,      'y', '<vis-register>+<vis-operator-yank>')
	vis:map(m.VISUAL,      'y', '<vis-register>+<vis-operator-yank>')
	vis:map(m.NORMAL,      'y', '<vis-register>+<vis-operator-yank>')
	vis:map(m.VISUAL,      'd', '<vis-register>+<vis-operator-delete>')
	vis:map(m.VISUAL_LINE, 'd', '<vis-register>+<vis-operator-delete>')
	vis:map(m.VISUAL_LINE, 'd', '<vis-register>+<vis-operator-delete>')
	vis:map(m.NORMAL,      'p', '<vis-register>+<vis-put-after>')
	vis:map(m.VISUAL,      'p', '<vis-register>+<vis-put-after>')
	vis:map(m.VISUAL_LINE, 'p', '<vis-register>+<vis-put-after>')
	vis:map(m.NORMAL,      'P', '<vis-register>+<vis-put-before>')
	vis:map(m.VISUAL,      'P', '<vis-register>+<vis-put-before>')
	vis:map(m.VISUAL_LINE, 'P', '<vis-register>+<vis-put-before>')

	vis:map(m.NORMAL, ' w', function() vis:command('w') end)
	vis:map(m.NORMAL, ' q', function() vis:command('q') end)
	vis:map(m.NORMAL, ' Q', function() vis:command('q!') end)

	vis:map(m.NORMAL, ' e', function()
		local s, f = vis:pipe('find -type f | vis-menu -l 5 -p "Edit file"')
		if s ~= 0 or not f then return end
		cmd = ('e "%s"'):format(f:sub(1, -2))
		vis:info(cmd)
		vis:command(cmd)
	end, 'Edit file')

	vis:map(m.NORMAL, ' cd', function()
		local s, f = vis:pipe('find -type d | vis-menu -l 5 -p "Change directory"')
		if s ~= 0 or not f then return end
		cmd = ('cd "%s"'):format(f:sub(1, -2))
		vis:info(cmd)
		vis:command(cmd)
	end, 'Change directory')

	-- editor options
	vis.options.autoindent = true
end)

-- window open
e.subscribe(e.WIN_OPEN, function(win)
	win.options.colorcolumn = 81
	win.options.numbers = true

	if win.syntax == 'markdown' then
		win.options.breakat = " ,]_"
		win.options.expandtab = true
		win.options.tabwidth = 2
		win.options.wrapcolumn = 81
	end

	if win.syntax == 'css' then
		win.options.expandtab = true
		win.options.tabwidth = 2
	end
end)

-- statusline
e.subscribe(e.WIN_STATUS, function(win)
	local left_parts = {}
	local right_parts = {}
	local file = win.file
	local selection = win.selection

	-- file info
	table.insert(left_parts,
		(file.name or '[No Name]')..(file.modified and '[+]' or ''))

	-- selection
	table.insert(right_parts, selection.number..'/'..#win.selections)

	if vis.win == win then
		-- mode
		table.insert(left_parts, 1, mode_strings[vis.mode])

		-- syntax
		table.insert(left_parts, win.syntax)

		-- input info
		table.insert(left_parts,
			'<'
			..(vis.count or '')
			..(vis.input_queue or '')
			..(vis.recording and '@' or '')
			..'>')

		-- character under cursor
		table.insert(right_parts,
			'<'
			..(string.byte(file:content(selection.pos, 1)) or '0')
			..'>')

		-- line and column count
		table.insert(right_parts, #file.lines..'/'..selection.line)
		table.insert(right_parts, selection.col)
	end

	-- fillchars
	local left = table.concat(left_parts, ' ')
	local right = table.concat(right_parts, ' ')

	win:status(table.concat({ left, ('^'):rep(win.width - #left - #right - 2)}, ' '), right)
end)

-- set title
local modified = false
local function set_title(title)
	os.execute('printf "\\e];'..title..(modified and '[+]' or '')..'\\e"')
end

e.subscribe(e.WIN_OPEN, function(win)
	set_title(win.file.name or '[No Name]')
end)

e.subscribe(e.FILE_SAVE_POST, function(file)
	modified = false
	set_title(file.name)
end)

e.subscribe(e.WIN_STATUS, function(win)
	if not modified and win.file.modified then
		modified = true
		set_title(win.file.name or '[No Name]')
	end
end)

-- find root
e.subscribe(e.WIN_OPEN, function(win)
	if not win.file.path then return end
	local dir = win.file.path
	local home = os.getenv('HOME')
	if not dir:find(home) then return end
	while true do
		dir = dir:match('^(.+)/[^/]+$') or home
		local _, find = vis:pipe(('find %s ! -path %s -prune -type d -name .git'):format(dir, dir))
		if find or dir == home then
			local cmd = ('cd "%s"'):format(dir)
			vis:info(cmd)
			vis:command(cmd)
			break
		end
	end
end)
